@@ -30,14 +30,14 @@
 
 
 // defines available
-// enable LFB support (depends upon bochs-vbe-lfb patch)
+// enable LFB support
 #define VBE_HAVE_LFB
 
 // disable VESA/VBE2 check in vbe info
 //#define VBE2_NO_VESA_CHECK
 
 // dynamicly generate a mode_info list
-//#define DYN_LIST
+#define DYN_LIST
 
 
 #include "vbe.h"
@@ -47,16 +47,11 @@
 // The current OEM Software Revision of this VBE Bios
 #define VBE_OEM_SOFTWARE_REV 0x0002;
 
-#define VBEInfoData ((VbeInfoBlock *) 0)
-
 extern char vbebios_copyright;
 extern char vbebios_vendor_name;
 extern char vbebios_product_name;
 extern char vbebios_product_revision;
 
-// FIXME: find out why we cannot use the dynamic list generation (due to a bug somewhere)
-//#define DYN_LIST
-
 #ifndef DYN_LIST
 extern Bit16u vbebios_mode_list;
 #endif
@@ -64,7 +59,7 @@ extern Bit16u vbebios_mode_list;
 ASM_START
 // FIXME: 'merge' these (c) etc strings with the vgabios.c strings?
 _vbebios_copyright:
-.ascii       "Bochs/Plex86 VBE(C) 2003 Jeroen Janssen <japj@darius.demon.nl>"
+.ascii       "Bochs/Plex86 VBE(C) 2003 http://savannah.nongnu.org/projects/vgabios/"
 .byte        0x00
 
 _vbebios_vendor_name:
@@ -76,7 +71,7 @@ _vbebios_product_name:
 .byte        0x00
 
 _vbebios_product_revision:
-.ascii       "$Id: vbe.c,v 1.35 2003/11/03 20:57:01 vruppert Exp $"
+.ascii       "$Id: vbe.c,v 1.45 2004/07/22 18:37:29 vruppert Exp $"
 .byte        0x00
 
 _vbebios_info_string:
@@ -91,6 +86,10 @@ _no_vbebios_info_string:
 .byte	0x0a,0x0d
 .byte 0x00
 
+msg_vbe_init:
+.ascii      "VBE Bios $Id: vbe.c,v 1.45 2004/07/22 18:37:29 vruppert Exp $"
+.byte	0x0a,0x0d, 0x00
+
 
 #ifndef DYN_LIST
 // FIXME: for each new mode add a statement here
@@ -134,28 +133,66 @@ MACRO HALT
   mov ax,#?1
   out dx,ax 
 MEND
+
+; DISPI ioport functions
+
+dispi_get_id:
+  push dx
+  mov  dx, # VBE_DISPI_IOPORT_INDEX
+  mov  ax, # VBE_DISPI_INDEX_ID
+  out  dx, ax
+  mov  dx, # VBE_DISPI_IOPORT_DATA
+  in   ax, dx
+  pop  dx
+  ret
+
+dispi_set_id:
+  push dx
+  push ax
+  mov  dx, # VBE_DISPI_IOPORT_INDEX
+  mov  ax, # VBE_DISPI_INDEX_ID
+  out  dx, ax
+  pop  ax
+  mov  dx, # VBE_DISPI_IOPORT_DATA
+  out  dx, ax
+  pop  dx
+  ret
 ASM_END
 
-// DISPI ioport functions
-// FIXME: what if no VBE host side code?
-static Bit16u dispi_get_id()
-{
-  outw(VBE_DISPI_IOPORT_INDEX,VBE_DISPI_INDEX_ID);
-  return inw(VBE_DISPI_IOPORT_DATA);  
-}
-
-static void dispi_set_id(id)
-  Bit16u id;
-{
-  outw(VBE_DISPI_IOPORT_INDEX,VBE_DISPI_INDEX_ID);
-  outw(VBE_DISPI_IOPORT_DATA,id);
-}
-
 static void dispi_set_xres(xres)
   Bit16u xres;
 {
-  outw(VBE_DISPI_IOPORT_INDEX,VBE_DISPI_INDEX_XRES);
-  outw(VBE_DISPI_IOPORT_DATA,xres);
+ASM_START
+  push bp
+  mov  bp, sp
+  push ax
+  push dx
+
+  mov  dx, # VBE_DISPI_IOPORT_INDEX
+  mov  ax, # VBE_DISPI_INDEX_XRES
+  out  dx, ax
+  mov  dx, # VBE_DISPI_IOPORT_DATA
+  mov  ax, 4[bp] ; xres
+  out  dx, ax
+  push ax
+  mov  dx, #0x03d4
+  mov  ax, #0x0011
+  out  dx, ax
+  mov  dx, #0x03d4
+  pop  ax
+  push ax
+  shr  ax, #3
+  dec  ax
+  mov  ah, al
+  mov  al, #0x01
+  out  dx, ax
+  pop  ax
+  call vga_set_virt_width
+
+  pop  dx
+  pop  ax
+  pop  bp
+ASM_END
 }
 
 static void dispi_set_yres(yres)
@@ -172,37 +209,88 @@ static void dispi_set_bpp(bpp)
   outw(VBE_DISPI_IOPORT_DATA,bpp);
 }
 
-static Bit16u dispi_get_bpp()
-{
-  outw(VBE_DISPI_IOPORT_INDEX,VBE_DISPI_INDEX_BPP);
-  return inw(VBE_DISPI_IOPORT_DATA);
-}
+ASM_START
+; AL = bits per pixel / AH = bytes per pixel
+dispi_get_bpp:
+  push dx
+  mov  dx, # VBE_DISPI_IOPORT_INDEX
+  mov  ax, # VBE_DISPI_INDEX_BPP
+  out  dx, ax
+  mov  dx, # VBE_DISPI_IOPORT_DATA
+  in   ax, dx
+  mov  ah, al
+  shr  ah, 3
+  test al, #0x07
+  jz   get_bpp_noinc
+  inc  ah
+get_bpp_noinc:
+  pop  dx
+  ret
 
-static Bit16u dispi_get_enable()
-{
-  outw(VBE_DISPI_IOPORT_INDEX,VBE_DISPI_INDEX_ENABLE);
-  return inw(VBE_DISPI_IOPORT_DATA);  
-}
+_dispi_get_max_bpp:
+  push dx
+  push bx
+  call dispi_get_enable
+  mov  bx, ax
+  or   ax, # VBE_DISPI_GETCAPS
+  call _dispi_set_enable
+  mov  dx, # VBE_DISPI_IOPORT_INDEX
+  mov  ax, # VBE_DISPI_INDEX_BPP
+  out  dx, ax
+  mov  dx, # VBE_DISPI_IOPORT_DATA
+  in   ax, dx
+  push ax
+  mov  ax, bx
+  call _dispi_set_enable
+  pop  ax
+  pop  bx
+  pop  dx
+  ret
 
-void dispi_set_enable(enable)
-  Bit16u enable;
-{
-  outw(VBE_DISPI_IOPORT_INDEX,VBE_DISPI_INDEX_ENABLE);
-  outw(VBE_DISPI_IOPORT_DATA,enable);
-}
+_dispi_set_enable:
+  push dx
+  push ax
+  mov  dx, # VBE_DISPI_IOPORT_INDEX
+  mov  ax, # VBE_DISPI_INDEX_ENABLE
+  out  dx, ax
+  pop  ax
+  mov  dx, # VBE_DISPI_IOPORT_DATA
+  out  dx, ax
+  pop  dx
+  ret
 
-static void dispi_set_bank(bank)
-  Bit16u bank;
-{
-  outw(VBE_DISPI_IOPORT_INDEX,VBE_DISPI_INDEX_BANK);
-  outw(VBE_DISPI_IOPORT_DATA,bank);
-}
+dispi_get_enable:
+  push dx
+  mov  dx, # VBE_DISPI_IOPORT_INDEX
+  mov  ax, # VBE_DISPI_INDEX_ENABLE
+  out  dx, ax
+  mov  dx, # VBE_DISPI_IOPORT_DATA
+  in   ax, dx
+  pop  dx
+  ret
 
-static Bit16u dispi_get_bank()
-{
-  outw(VBE_DISPI_IOPORT_INDEX,VBE_DISPI_INDEX_BANK);
-  return inw(VBE_DISPI_IOPORT_DATA);
-}
+_dispi_set_bank:
+  push dx
+  push ax
+  mov  dx, # VBE_DISPI_IOPORT_INDEX
+  mov  ax, # VBE_DISPI_INDEX_BANK
+  out  dx, ax
+  pop  ax
+  mov  dx, # VBE_DISPI_IOPORT_DATA
+  out  dx, ax
+  pop  dx
+  ret
+
+dispi_get_bank:
+  push dx
+  mov  dx, # VBE_DISPI_IOPORT_INDEX
+  mov  ax, # VBE_DISPI_INDEX_BANK
+  out  dx, ax
+  mov  dx, # VBE_DISPI_IOPORT_DATA
+  in   ax, dx
+  pop  dx
+  ret
+ASM_END
 
 static void dispi_set_bank_farcall()
 {
@@ -233,57 +321,104 @@ dispi_set_bank_farcall_error:
 ASM_END
 }
 
-static void dispi_set_x_offset(offset)
-  Bit16u offset;
-{
-  outw(VBE_DISPI_IOPORT_INDEX,VBE_DISPI_INDEX_X_OFFSET);
-  outw(VBE_DISPI_IOPORT_DATA,offset);
-}
+ASM_START
+dispi_set_x_offset:
+  push dx
+  push ax
+  mov  dx, # VBE_DISPI_IOPORT_INDEX
+  mov  ax, # VBE_DISPI_INDEX_X_OFFSET
+  out  dx, ax
+  pop  ax
+  mov  dx, # VBE_DISPI_IOPORT_DATA
+  out  dx, ax
+  pop  dx
+  ret
 
-static Bit16u dispi_get_x_offset()
-{
-  outw(VBE_DISPI_IOPORT_INDEX,VBE_DISPI_INDEX_X_OFFSET);
-  return inw(VBE_DISPI_IOPORT_DATA);
-}
+dispi_get_x_offset:
+  push dx
+  mov  dx, # VBE_DISPI_IOPORT_INDEX
+  mov  ax, # VBE_DISPI_INDEX_X_OFFSET
+  out  dx, ax
+  mov  dx, # VBE_DISPI_IOPORT_DATA
+  in   ax, dx
+  pop  dx
+  ret
 
-static void dispi_set_y_offset(offset)
-  Bit16u offset;
-{
-  outw(VBE_DISPI_IOPORT_INDEX,VBE_DISPI_INDEX_Y_OFFSET);
-  outw(VBE_DISPI_IOPORT_DATA,offset);
-}
+dispi_set_y_offset:
+  push dx
+  push ax
+  mov  dx, # VBE_DISPI_IOPORT_INDEX
+  mov  ax, # VBE_DISPI_INDEX_Y_OFFSET
+  out  dx, ax
+  pop  ax
+  mov  dx, # VBE_DISPI_IOPORT_DATA
+  out  dx, ax
+  pop  dx
+  ret
 
-static Bit16u dispi_get_y_offset()
-{
-  outw(VBE_DISPI_IOPORT_INDEX,VBE_DISPI_INDEX_Y_OFFSET);
-  return inw(VBE_DISPI_IOPORT_DATA);
-}
+dispi_get_y_offset:
+  push dx
+  mov  dx, # VBE_DISPI_IOPORT_INDEX
+  mov  ax, # VBE_DISPI_INDEX_Y_OFFSET
+  out  dx, ax
+  mov  dx, # VBE_DISPI_IOPORT_DATA
+  in   ax, dx
+  pop  dx
+  ret
 
-static void dispi_set_virt_width(width)
-  Bit16u width;
-{
-  outw(VBE_DISPI_IOPORT_INDEX,VBE_DISPI_INDEX_VIRT_WIDTH);
-  outw(VBE_DISPI_IOPORT_DATA,width);
-}
+vga_set_virt_width:
+  push ax
+  push bx
+  push dx
+  mov  bx, ax
+  call dispi_get_bpp
+  cmp  al, #0x04
+  ja   set_width_svga
+  shr  bx, #2
+set_width_svga:
+  shr  bx, #2
+  mov  dx, #0x03d4
+  mov  ah, bl
+  mov  al, #0x13
+  out  dx, ax
+  pop  dx
+  pop  bx
+  pop  ax
+  ret
 
-static Bit16u dispi_get_virt_width()
-{
-  outw(VBE_DISPI_IOPORT_INDEX,VBE_DISPI_INDEX_VIRT_WIDTH);
-  return inw(VBE_DISPI_IOPORT_DATA);
-}
-/*
-static void dispi_set_virt_height(height)
-  Bit16u height;
-{
-  outw(VBE_DISPI_IOPORT_INDEX,VBE_DISPI_INDEX_VIRT_HEIGHT);
-  outw(VBE_DISPI_IOPORT_DATA,height);
-}
-*/
-static Bit16u dispi_get_virt_height()
-{
-  outw(VBE_DISPI_IOPORT_INDEX,VBE_DISPI_INDEX_VIRT_HEIGHT);
-  return inw(VBE_DISPI_IOPORT_DATA);
-}
+dispi_set_virt_width:
+  call vga_set_virt_width
+  push dx
+  push ax
+  mov  dx, # VBE_DISPI_IOPORT_INDEX
+  mov  ax, # VBE_DISPI_INDEX_VIRT_WIDTH
+  out  dx, ax
+  pop  ax
+  mov  dx, # VBE_DISPI_IOPORT_DATA
+  out  dx, ax
+  pop  dx
+  ret
+
+dispi_get_virt_width:
+  push dx
+  mov  dx, # VBE_DISPI_IOPORT_INDEX
+  mov  ax, # VBE_DISPI_INDEX_VIRT_WIDTH
+  out  dx, ax
+  mov  dx, # VBE_DISPI_IOPORT_DATA
+  in   ax, dx
+  pop  dx
+  ret
+
+dispi_get_virt_height:
+  push dx
+  mov  dx, # VBE_DISPI_IOPORT_INDEX
+  mov  ax, # VBE_DISPI_INDEX_VIRT_HEIGHT
+  out  dx, ax
+  mov  dx, # VBE_DISPI_IOPORT_DATA
+  in   ax, dx
+  pop  dx
+  ret
+ASM_END
 
 
 // ModeInfo helper function
@@ -318,66 +453,67 @@ static ModeInfoListItem* mode_info_find_mode(mode, using_lfb)
   return 0;
 }
 
-/** Has VBE display - Returns true if VBE display detected
- *
- */
-Boolean vbe_has_vbe_display()
-{
-  dispi_set_id(VBE_DISPI_ID2);
+ASM_START
 
-  return (dispi_get_id()==VBE_DISPI_ID2);
-}
+; Has VBE display - Returns true if VBE display detected
 
-/** VBE Init - Initialise the Vesa Bios Extension Code
- *
- *  This function does a sanity check on the host side display code interface.
- */
-void vbe_init()
-{
-  Bit16u dispi_id;
-  
-  outw(VBE_DISPI_IOPORT_INDEX,VBE_DISPI_INDEX_ID);
-  outw(VBE_DISPI_IOPORT_DATA,VBE_DISPI_ID0);
-  
-  dispi_id=inw(VBE_DISPI_IOPORT_DATA);
-  
-  if (dispi_id!=VBE_DISPI_ID0)
-  {
-//FIXME this results in a 'rombios.c' line panic, but it's actually a 'vbe.c' panic
-ASM_START    
-    HALT(__LINE__)
-ASM_END    
-  }
-//#ifdef DEBUG
-  printf("VBE Bios $Id: vbe.c,v 1.35 2003/11/03 20:57:01 vruppert Exp $\n");
-//#endif  
-}
+_vbe_has_vbe_display:
+  push ds
+  push bx
+  mov  ax, # BIOSMEM_SEG
+  mov  ds, ax
+  mov  bx, # BIOSMEM_VBE_FLAG
+  mov  al, [bx]
+  and  al, #0x01
+  xor  ah, ah
+  pop  bx
+  pop  ds
+  ret
 
-/** VBE Display Info - Display information on screen about the VBE
- */
-void vbe_display_info()
-{
-  // Check for VBE display extension in Bochs
-  if (vbe_has_vbe_display())
-  {
-    ASM_START
-     mov ax,#0xc000
-     mov ds,ax
-     mov si,#_vbebios_info_string
-     call _display_string
-    ASM_END  
-  }
-  else
-  {
-    ASM_START
-     mov ax,#0xc000
-     mov ds,ax
-     mov si,#_no_vbebios_info_string
-     call _display_string
-    ASM_END  
-    
-  }
-}  
+; VBE Init - Initialise the Vesa Bios Extension Code
+; This function does a sanity check on the host side display code interface.
+
+vbe_init:
+  mov  ax, # VBE_DISPI_ID0
+  call dispi_set_id
+  call dispi_get_id
+  cmp  ax, # VBE_DISPI_ID0
+  jne  no_vbe_interface
+  push ds
+  push bx
+  mov  ax, # BIOSMEM_SEG
+  mov  ds, ax
+  mov  bx, # BIOSMEM_VBE_FLAG
+  mov  al, #0x01
+  mov  [bx], al
+  pop  bx
+  pop  ds
+  mov  ax, # VBE_DISPI_ID3
+  call dispi_set_id
+no_vbe_interface:
+  mov  bx, #msg_vbe_init
+  push bx
+  call _printf
+  inc  sp
+  inc  sp
+  ret
+
+; VBE Display Info - Display information on screen about the VBE
+
+vbe_display_info:
+  call _vbe_has_vbe_display
+  test ax, ax
+  jz   no_vbe_flag
+  mov  ax, #0xc000
+  mov  ds, ax
+  mov  si, #_vbebios_info_string
+  jmp  _display_string
+no_vbe_flag:
+  mov  ax, #0xc000
+  mov  ds, ax
+  mov  si, #_no_vbebios_info_string
+  jmp  _display_string
+ASM_END  
 
 /** Function 00h - Return VBE Controller Information
  * 
@@ -398,10 +534,8 @@ Bit16u *AX;Bit16u ES;Bit16u DI;
         Bit16u            status;
         Bit16u            result;
         Bit16u            vbe2_info;
-#ifdef DYN_LIST
-        Bit16u            *video_mode_list;
-#endif
         Bit16u            cur_mode=0;
+        Bit16u            cur_ptr=34;
         ModeInfoListItem  *cur_info=&mode_info_list;
         
         status = read_word(ss, AX);
@@ -448,42 +582,19 @@ Bit16u *AX;Bit16u ES;Bit16u DI;
         vbe_info_block.OemStringPtr_Off = &vbebios_copyright;
         
         // Capabilities
-        vbe_info_block.Capabilities[0] = 0;
+        vbe_info_block.Capabilities[0] = VBE_CAPABILITY_8BIT_DAC;
         vbe_info_block.Capabilities[1] = 0;
         vbe_info_block.Capabilities[2] = 0;
         vbe_info_block.Capabilities[3] = 0;
 
 #ifdef DYN_LIST
-        // FIXME: This doesn't work correctly somehow?
         // VBE Video Mode Pointer (dynamicly generated from the mode_info_list)
-        vbe_info_block.VideoModePtr_Seg= ES ;//0xc000;
-        vbe_info_block.VideoModePtr_Off= DI + 34;//&(VBEInfoData->Reserved);//&vbebios_mode_list;
+        vbe_info_block.VideoModePtr_Seg= ES ;
+        vbe_info_block.VideoModePtr_Off= DI + 34;
 #else
         // VBE Video Mode Pointer (staticly in rom)
         vbe_info_block.VideoModePtr_Seg = 0xc000;
         vbe_info_block.VideoModePtr_Off = &vbebios_mode_list;
-
-#endif
-
-#ifdef DYN_LIST
-
-//      video_mode_list=(Bit16u*)&(vbe_info_block.Reserved);
-
-        do
-        {
-#ifdef DEBUG
-                printf("VBE found mode %x => %x\n", cur_info->mode,cur_mode);
-#endif
-//              *video_mode_list=cur_info->mode;
-                vbe_info_block.Reserved[cur_mode] = cur_info->mode;
-                
-                cur_info++;
-                //video_mode_list++;
-                cur_mode++;
-        } while (cur_info->mode != VBE_VESA_MODE_END_OF_LIST);
-        
-        // Add vesa mode list terminator
-        vbe_info_block.Reserved[cur_mode] = VBE_VESA_MODE_END_OF_LIST;
 #endif
 
         // VBE Total Memory (in 64b blocks)
@@ -509,6 +620,24 @@ Bit16u *AX;Bit16u ES;Bit16u DI;
                 memcpyb(ES, DI, ss, &vbe_info_block, 256);
 	}
                 
+#ifdef DYN_LIST
+        do
+        {
+                if (cur_info->info.BitsPerPixel <= dispi_get_max_bpp()) {
+#ifdef DEBUG
+                  printf("VBE found mode %x => %x\n", cur_info->mode,cur_mode);
+#endif
+                  write_word(ES, DI + cur_ptr, cur_info->mode);
+                  cur_mode++;
+                  cur_ptr+=2;
+                }
+                cur_info++;
+        } while (cur_info->mode != VBE_VESA_MODE_END_OF_LIST);
+        
+        // Add vesa mode list terminator
+        write_word(ES, DI + cur_ptr, cur_info->mode);
+#endif
+
         result = 0x4f;
 
         write_word(ss, AX, result);
@@ -637,9 +766,9 @@ Bit16u *AX;Bit16u BX; Bit16u ES;Bit16u DI;
                   biosfn_set_video_mode(0x6a);
                 }
 
+                dispi_set_bpp(cur_info->info.BitsPerPixel);
                 dispi_set_xres(cur_info->info.XResolution);
                 dispi_set_yres(cur_info->info.YResolution);
-                dispi_set_bpp(cur_info->info.BitsPerPixel);
                 dispi_set_bank(0);
                 dispi_set_enable(VBE_DISPI_ENABLED | no_clear | lfb_flag);
 
@@ -673,27 +802,28 @@ Bit16u *AX;Bit16u BX; Bit16u ES;Bit16u DI;
  *              BX      = Current VBE Mode
  * 
  */
-void vbe_biosfn_return_current_mode(AX, BX)
-Bit16u *AX;Bit16u *BX;
-{
-        Bit16u ss=get_SS();
-        Bit16u mode;
-
-#ifdef DEBUG
-        printf("VBE vbe_biosfn_return_current_mode\n");
-#endif
-
-        if(dispi_get_enable())
-        {
-                mode=read_word(BIOSMEM_SEG,BIOSMEM_VBE_MODE);
-        }
-        else
-        {
-                mode=read_byte(BIOSMEM_SEG,BIOSMEM_CURRENT_MODE);
-        }
-        write_word(ss, AX, 0x4f);
-        write_word(ss, BX, mode);
-}
+ASM_START
+vbe_biosfn_return_current_mode:
+  push ds
+  mov  ax, # BIOSMEM_SEG
+  mov  ds, ax
+  call dispi_get_enable
+  and  ax, # VBE_DISPI_ENABLED
+  jz   no_vbe_mode
+  mov  bx, # BIOSMEM_VBE_MODE
+  mov  ax, [bx]
+  mov  bx, ax
+  jnz  vbe_03_ok
+no_vbe_mode:
+  mov  bx, # BIOSMEM_CURRENT_MODE
+  mov  al, [bx]
+  mov  bl, al
+  xor  bh, bh
+vbe_03_ok:
+  mov  ax, #0x004f
+  pop  ds
+  ret
+ASM_END
 
 
 /** Function 04h - Save/Restore State
@@ -735,26 +865,32 @@ void vbe_biosfn_save_restore_state(AX, DL, CX, ES, BX)
  *              DX      = Window number in window granularity units
  *                        (Get Memory Window only)
  */
-void vbe_biosfn_display_window_control(AX,BX,DX)
-Bit16u *AX;Bit16u BX;Bit16u *DX;
-{
-        Bit16u            ss = get_SS();
-        Bit16u            window = read_word(ss, DX);
-        Bit16u            result = 0x014f;
-        
-        if (BX==0x0000)
-        {
-                dispi_set_bank(window);
-                result = 0x4f;
-        }
-        else if (BX==0x0100)
-        {
-                window = dispi_get_bank();
-                write_word(ss, DX, result);
-                result = 0x4f;
-        }
-        write_word(ss, AX, result);
-}
+ASM_START
+vbe_biosfn_display_window_control:
+  cmp  bl, #0x00
+  jne  vbe_05_failed
+  cmp  bh, #0x01
+  je   get_display_window
+  jb   set_display_window
+  mov  ax, #0x0100
+  ret
+set_display_window:
+  mov  ax, dx
+  call _dispi_set_bank
+  call dispi_get_bank
+  cmp  ax, dx
+  jne  vbe_05_failed
+  mov  ax, #0x004f
+  ret
+get_display_window:
+  call dispi_get_bank
+  mov  dx, ax
+  mov  ax, #0x004f
+  ret
+vbe_05_failed:
+  mov  ax, #0x014f
+  ret
+ASM_END
 
 
 /** Function 06h - Set/Get Logical Scan Line Length
@@ -776,51 +912,39 @@ Bit16u *AX;Bit16u BX;Bit16u *DX;
  *                        (truncated to nearest complete pixel)
  *              DX      = Maximum Number of Scan Lines 
  */
-void vbe_biosfn_set_get_logical_scan_line_length(AX,BX,CX,DX)
-Bit16u *AX;Bit16u *BX;Bit16u *DX;Bit16u *DX;
-{
-	Bit16u ss=get_SS();
-	Bit16u result=0x100;
-	Bit16u width = read_word(ss, CX);
-	Bit16u cmd = read_word(ss, BX);
-	Bit8u  bytespp = dispi_get_bpp()/8;
-
-	// check bl
-	if ( ((cmd & 0xff) == 0x00) || ((cmd & 0xff) == 0x02) )
-	{
-		// set scan line lenght in pixels(0x00) or bytes (0x02)
-		Bit16u new_width;
-		Bit16u new_height;
-
-		if ( ((cmd & 0xff) == 0x02) && (bytespp > 1) )
-		{
-			width/=bytespp;
-		}
-		dispi_set_virt_width(width);
-		new_width=dispi_get_virt_width();
-		new_height=dispi_get_virt_height();
-		
-		if (new_width!=width)
-		{
-#ifdef DEBUG
-                printf("* VBE width adjusted\n");
-#endif
-			
-			// notify width adjusted
-			result=0x024f;
-		}
-		else
-		{
-			result=0x4f;
-		}
-		
-		write_word(ss,BX,new_width*bytespp);
-		write_word(ss,CX,new_width);
-		write_word(ss,DX,new_height);
-	}
-
-	write_word(ss, AX, result);	
-}
+ASM_START
+vbe_biosfn_set_get_logical_scan_line_length:
+  mov  ax, cx
+  cmp  bl, #0x01
+  je   get_logical_scan_line_length
+  cmp  bl, #0x02
+  je   set_logical_scan_line_bytes
+  jb   set_logical_scan_line_pixels
+  mov  ax, #0x0100
+  ret
+set_logical_scan_line_bytes:
+  push ax
+  call dispi_get_bpp
+  xor  bh, bh
+  mov  bl, ah
+  xor  dx, dx
+  pop  ax
+  div  bx
+set_logical_scan_line_pixels:
+  call dispi_set_virt_width
+get_logical_scan_line_length:
+  call dispi_get_bpp
+  xor  bh, bh
+  mov  bl, ah
+  call dispi_get_virt_width
+  mov  cx, ax
+  mul  bx
+  mov  bx, ax
+  call dispi_get_virt_height
+  mov  dx, ax
+  mov  ax, #0x004f
+  ret
+ASM_END
 
 
 /** Function 07h - Set/Get Display Start
@@ -859,44 +983,76 @@ Bit16u *AX;Bit16u *BX;Bit16u *DX;Bit16u *DX;
  *              DX      = Bits 16-31 of display start address
  *              ES      = Selector for memory mapped registers 
  */
-void vbe_biosfn_set_get_display_start(AX,BX,CX,DX)
-Bit16u *AX;Bit16u BX;Bit16u CX;Bit16u DX;
-{
-	Bit16u ss=get_SS();
-	Bit16u result=0x100;
-#ifdef DEBUG
- //       printf("VBE vbe_biosfn_set_get_display_start\n");
-#endif
-	
-	// check for set display start
-	if ((GET_BL()==0x00) || (GET_BL()==0x80))
-	{
-	  // 0x80 is during vertical retrace - is also used by sdd vbetest during multibuffering
-#ifdef DEBUG
-//        printf("VBE vbe_biosfn_set_get_display_start CX%x DX%x\n",CX,DX);
-#endif
-		
-		dispi_set_x_offset(CX);
-		dispi_set_y_offset(DX);
-		result = 0x4f;
-	}
-	
-	write_word(ss, AX, result);	
-}
-
+ASM_START
+vbe_biosfn_set_get_display_start:
+  cmp  bl, #0x80
+  je   set_display_start
+  cmp  bl, #0x01
+  je   get_display_start
+  jb   set_display_start
+  mov  ax, #0x0100
+  ret
+set_display_start:
+  mov  ax, cx
+  call dispi_set_x_offset
+  mov  ax, dx
+  call dispi_set_y_offset
+  mov  ax, #0x004f
+  ret
+get_display_start:
+  call dispi_get_x_offset
+  mov  cx, ax
+  call dispi_get_y_offset
+  mov  dx, ax
+  xor  bh, bh
+  mov  ax, #0x004f
+  ret
+ASM_END
+  
 
 /** Function 08h - Set/Get Dac Palette Format
  * 
  * Input:
  *              AX      = 4F08h
+ *              BL      = 00h set DAC palette width
+ *                      = 01h get DAC palette width
+ *              BH      = If BL=00h: desired number of bits per primary color
  * Output:
  *              AX      = VBE Return Status
- *
- * FIXME: incomplete API description, Input & Output
+ *              BH      = current number of bits per primary color (06h = standard VGA)
  */
-void vbe_biosfn_set_get_dac_palette_format(AX)
-{
-}
+ASM_START
+vbe_biosfn_set_get_dac_palette_format:
+  cmp  bl, #0x01
+  je   get_dac_palette_format
+  jb   set_dac_palette_format
+  mov  ax, #0x0100
+  ret
+set_dac_palette_format:
+  call dispi_get_enable
+  cmp  bh, #0x06
+  je   set_normal_dac
+  cmp  bh, #0x08
+  jne  vbe_08_unsupported
+  or   ax, # VBE_DISPI_8BIT_DAC
+  jnz  set_dac_mode
+set_normal_dac:
+  and  ax, #~ VBE_DISPI_8BIT_DAC
+set_dac_mode:
+  call _dispi_set_enable
+get_dac_palette_format:
+  mov  bh, #0x06
+  call dispi_get_enable
+  and  ax, # VBE_DISPI_8BIT_DAC
+  jz   vbe_08_ok
+  mov  bh, #0x08
+vbe_08_ok:
+  mov  ax, #0x004f
+  ret
+vbe_08_unsupported:
+  mov  ax, #0x014f
+  ret
+ASM_END
 
 
 /** Function 09h - Set/Get Palette Data
